#======
# Copyright (c) IBM Corp. 2011, 2015.
# All rights reserved. This program is made available
# under the terms of the Eclipse Public License v1.0
# which accompanies this distribution, and is available at
# http://www.eclipse.org/legal/epl-v10.html
#=======


from pykickstart.version import returnClassForVersion

from controller.pkvmerror import PKVMError
from model.config import PRESERVABLE_MOUNT_POINTS
from modules.installauto.pkvmkickstart.commands.logvol import Pkvm_LogVol, Pkvm_LogVolData
from modules.installauto.pkvmkickstart.commands.partition import Pkvm_Partition
from modules.installauto.pkvmkickstart.commands.volgroup import Pkvm_VolGroup


PKVM_KS_ACCEPTED_CMDS=['lang', 'logvol', 'network', 'part', 'partition',
                       'rootpw', 'timezone', 'volgroup']


# With no arguments, means DEVEL as defined in pykickstart.version
superclass = returnClassForVersion()

class PkvmHandler(superclass):
    # Override the behaviour of the following commands
    commandMap = {
        "logvol": Pkvm_LogVol,
        "part": Pkvm_Partition, "partition": Pkvm_Partition,
        "volgroup": Pkvm_VolGroup
    }

    dataMap = {
        "LogVolData": Pkvm_LogVolData
    }

    def __init__(self, mapping={}):
        """
        Args:
            mapping (dict): classes that handle each command
                Dictionary format:
                    { str : KickstartCommand }
        """

        superclass.__init__(self, mapping=mapping,
                            commandUpdates=self.commandMap,
                            dataUpdates=self.dataMap)
        self.maskAllExcept(PKVM_KS_ACCEPTED_CMDS) # pylint: disable=no-member
        self.formatAllLVs = None

    def validate(self):
        """
        Any extra validity checks pertaining to the currently supported syntax
        """
        # pylint: disable=no-member
        if not self.partition.seen == self.volgroup.seen:
            raise PKVMError("064")

        if self.volgroup.seen:
            if self.formatAllLVs is False:
                raise PKVMError("065")

            if len(self.volgroup.vgList) > 1:
                raise PKVMError("053")

            physicalVolumes = self.volgroup.vgList[0].physvols
            partitionNames = [part.mountpoint for part in
                              self.partition.partitions]

            if not set(physicalVolumes).issubset(set(partitionNames)):
                raise PKVMError("056")

            vgName = self.volgroup.vgList[0].vgname
            for lv in self.logvol.lvList:
                if lv.vgname != vgName:
                    errorMsgData = {"mountPoint": lv.mountpoint,
                                    "vgName": lv.vgname}
                    raise PKVMError("057", errorMsgData)

        else:
            if not self.logvol.seen:
                raise PKVMError("066")
            mountPoints = [lv.mountpoint for lv in self.logvol.lvList]
            if not self.__allPreservable(mountPoints):
                errorMsgData = {"numberOfPreservable": len(PRESERVABLE_MOUNT_POINTS)}
                raise PKVMError("062", errorMsgData)

    def __allPreservable(self, mountPoints):
        """
        Check if all the mount points are preservable

        Args:
            mountPoints ([str]): mount points

        Returns:
            bool: whether all mount points are preservable
        """

        for lvs in PRESERVABLE_MOUNT_POINTS.values():
            if len(mountPoints) == len(lvs) and set(mountPoints) == set(lvs):
                return True
        return False
