#======
# Copyright (c) IBM Corp. 2011, 2015.
# All rights reserved. This program is made available
# under the terms of the Eclipse Public License v1.0
# which accompanies this distribution, and is available at
# http://www.eclipse.org/legal/epl-v10.html
#=======


"""This scripts provides the partitioner for PowerKVM."""

#
# IMPORTS
#
import glob
import logging
import os
import pprint

from blivet import Size
from blivet import udev
from blivet.devices import PartitionDevice
from blivet.util import mount, run_program

from .storage import Storage
from controller.pkvmerror import PKVMError
from model.config import BOOT_SIZE
from model.config import LVS_METADATA
from model.config import LV_PREFIX
from model.config import LV_SYSTEM
from model.config import LV_VAR
from model.config import PKVM2_LV_LOG
from model.config import PKVM2_LV_ROOT
from model.config import PREP_SIZE
from modules.grub.grub import Grub

#
# CONSTANTS
#
SYSTEMD_TIMEOUT = 300
LVMSD = r's@^\(\s*filter =\).*@\1 [ "r/disk/", "r/sd.*/", "a/.*/" ]@'
MPTHSD = r's@add_dracutmodules.*@add_dracutmodules+=" systemd multipath "@'


#
# CODE
#
class Partitioner(object):
    """
    Implements the PowerKVM partitioner class.
    """

    def __init__(self):
        """
        Creates a partitioner.
        """
        self.__storage = Storage()

        self.__partitions = None
        self.__raids = None
        self.__vgs = None
        self.__lvs = None
        self.VG_ROOT = ""
        self.lvsBlivetNames = {}
        self.targetLVMInfo = None

        self.__mountpartitions = {}
        self.__mountpoint = None

        self.__logger = logging.getLogger(__name__)


    def __isSAN(self, path):
        """
        Checks if the disk is a fibre-channel.

        Args:
            disk (str): disk syspath

        Returns:
            bool: True if is fibre-channel
        """
        # get the full device path
        linkpath = '/sys/block/%s' % path
        realpath = os.path.realpath(linkpath)

        # realpath should not be the same because we
        # need the san disk path in order to verify if
        # it's fibre-channel
        if realpath == linkpath:
            return None

        # as per udev (udev-builtin-path_id.c), rport means the disk
        # is a fibre-channel
        if '/rport-' in realpath:
            return True

        return False


    def _getStorage(self):
        """
        Returns the storage object for arbitrary handling.

        Returns:
            Storage: storage object
        """
        return self.__storage


    def getInstallableDisks(self):
        """
        Filters the devices for disks that the OS could be
        installed.

        Returns:
            [Disk]: list of disks
        """
        self.__logger.debug('getting disks available to install')
        disks = []
        devices = self.__storage.getDisks()

        for dev in devices:

            # ignore SAN disks
            if self.__isSAN(dev.sysfsPath) is True:
                self.__logger.warn('%s is a SAN disk', dev.name)
                continue

            if dev.isDisk is True and dev.partitionable is True:
                disks.append(dev)

            self.__logger.warn('%s was not considered', dev.name)

        return disks


    def getSystemInformation(self):
        """
        Returns system information about devices and partitions.

        IMPORTANT: the structure might not reflect the current system
        state because changes in blivet's queue will be reflected -
        even if it's not yet commited. If you want to be sure to get
        the current system state, call 'scanDisks' just before
        this one.

        Returns:
            dict: devices and partitions data
        """
        return self.__storage.getSystemInformation()


    def mountPartitions(self, mpoint):
        """
        Using the created partition layout, mount all partitions in
            mountpoint including system partitions, letting the device
            ready to be installed.

        Args:
            mpoint (str): root mount point
        """
        self.__logger.info('Mounting partitions in %s', mpoint)
        self.__mountpoint = mpoint

        # root must be mounted first
        self.__mountpartitions['/'].format.mount(mountpoint=mpoint)

        # mount all devices
        for key, item in self.__mountpartitions.iteritems():

            # ignore no valid path (like swap, prep and root)
            if key[0] != '/' or len(key) <= 1:
                continue

            item.format.mount(mountpoint=mpoint + key)

        # mount system devices
        mount('/dev', os.path.join(mpoint, 'dev'), '', 'bind')
        mount('/sys', os.path.join(mpoint, 'sys'), '', 'bind')
        mount('/proc', os.path.join(mpoint, 'proc'), '', 'bind')
        mount('/run', os.path.join(mpoint, 'run'), '', 'bind')
        mount('/dev/shm', os.path.join(mpoint, 'dev/shm'), '', 'bind')

        # do not break is selinux is disabled
        if os.path.isdir('/sys/fs/selinux'):
            mount('/sys/fs/selinux', os.path.join(mpoint, 'sys/fs/selinux'),
                  '', 'bind')


    def addMountPoint(self, mountpoint, device):
        """
        Add an existing partition in the structure to be used in the
            partitioner.

        Args:
            mountpoint (str): partition mount point
            device (Blivet:DiskDevice): device to be included
        """
        self.__mountpartitions[mountpoint] = device


    def getMountPoints(self):
        """
        Returns the path of each partition by its mountpoint.

        Returns:
            dict: partition path
        """
        ret = {}
        # get the fstab spec for each partition
        for key, item in self.__mountpartitions.iteritems():

            # fstabSpec is not returning the partition UUID as expected so
            # we force UUID for any mountable partition
            if isinstance(item, PartitionDevice) and item.format.mountable:
                info = udev.get_device(item.sysfsPath)
                ret[key] = 'UUID=%s' % udev.device_get_uuid(info)

            else:
                ret[key] = item.fstabSpec

        self.__logger.debug('fstab spec: %s', str(ret))

        return ret


    def updateFstab(self, fstabContent):
        """
        Update /etc/fstab file with the correct information about
            fresh installed system.

        Args:
            fstabContent (str): fstab template to be written in file
        """
        # create /etc if it doesn't exist yet
        etc = os.path.join(self.__mountpoint, 'etc')
        if not os.path.exists(etc):
            os.makedirs(etc)

        # write the fstab file
        with open(os.path.join(etc, 'fstab'), 'w') as fstab:
            fstab.write(fstabContent)


    def setupMultipath(self):
        """
        Configure multipath in the installed system.
        """

        # get system directories
        lvmconf = os.path.join(self.__mountpoint, 'etc/lvm/lvm.conf')
        mpathdir = os.path.join(self.__mountpoint, 'etc/multipath')

        # copy mpath config file
        for mfile in glob.glob('/etc/multipath/*'):
            cmd = ['cp', mfile, mpathdir]
            if run_program(argv=cmd) != 0:
                self.__logger.critical('File %d was not copied successfuly',
                                       mfile)
                raise PKVMError("016")

        # lvm must use mpath in place of /dev/sd*
        if self.__mountpartitions['prep'].parents[0].type == 'dm-multipath':
            cmd = ['sed',
                   '-i',
                   LVMSD,
                   lvmconf]
            if run_program(argv=cmd) != 0:
                self.__logger.critical('Multipath was not setup successfuly')
                raise PKVMError("074")

        # http://www.redhat.com/archives/lvm-devel/2012-July/msg00072.html
        # multipath.service needs to run before LVM, but it only happend when
        # lvmetad is 0 - thats because when lvmetad is 1 it will use lvm data
        # from cache, not scanned by service.
        cmd = ['sed', '-i', r's@^\(\s*use_lvmetad =\).*@\1 0@', lvmconf]
        if run_program(argv=cmd) != 0:
            self.__logger.critical('LVM was not configured successfuly')
            raise PKVMError("075")

        # enable multipath in the chroot
        cmd = ['mpathconf', '--enable', '--with_multipathd', 'y']
        if run_program(argv=cmd, root=self.__mountpoint) != 0:
            self.__logger.error('Multipath not configured successfuly')
            raise PKVMError("076")

        # configure and updates the initrd
        cmd = ['sed',
               '-i',
               MPTHSD,
               '%s/lib/dracut/dracut.conf.d/01-dist.conf' % self.__mountpoint]
        if run_program(argv=cmd) != 0:
            self.__logger.error('Dracut was not configured successfuly')
            raise PKVMError("077")

        cmd = ['dracut', '--force']
        if run_program(argv=cmd, root=self.__mountpoint) != 0:
            self.__logger.error('Dracut initrd creation has failed')
            raise PKVMError("078")


    def installPRePBootloader(self):
        """
        Install grub in the PReP partition of the installed system
        """
        prep = self.__mountpartitions['prep']

        # etc dir not created yet: do it
        etc = os.path.join(self.__mountpoint, 'etc')
        if not os.path.exists(etc):
            os.makedirs(etc)

        # create the customized grub default config file
        etcDefault = os.path.join(etc, 'default')
        if not os.path.isdir(etcDefault):
            os.makedirs(etcDefault)

        lvFullNames = [lv.name.replace(lv.vg.name + '-', lv.vg.name + '/')
                       for lv in self.targetLVMInfo['lvs']]
        grubData = ['rd.lvm.lv=' + lvFullName for lvFullName in lvFullNames]

        grub = Grub()
        grub.setDefaultOptions(os.path.join(etcDefault, 'grub'), grubData)

        # generate grub.conf in chroot
        cmd = ['grub2-mkconfig', '-o', '/boot/grub2/grub.cfg']
        if run_program(argv=cmd, root=self.__mountpoint) != 0:
            self.__logger.critical('grub.conf was not created successfuly')
            raise PKVMError("017")

        # create a symlink for grub2.cfg in /etc
        grubcfg = os.path.join(etc, 'grub2.cfg')
        if os.path.exists(grubcfg):
            os.remove(grubcfg)

        cmd = ['ln',
               '-s',
               '/boot/grub2/grub.cfg',
               os.path.join(etc, 'grub2.cfg')]
        if run_program(argv=cmd) != 0:
            self.__logger.critical('grub.conf was not linked successfuly')
            raise PKVMError("096")

        # clean prep partition - not critical since petitboot does not use it
        cmd = ['dd', 'if=/dev/zero', 'of=%s' % prep.path]
        if run_program(argv=cmd) != 0:
            self.__logger.error('PReP was not cleaned accordingly')

        # install grub in prep partition - not critical to petitboot
        cmd = ['grub2-install', '--no-nvram', '--no-floppy', prep.path]
        if run_program(argv=cmd, root=self.__mountpoint):
            self.__logger.error('grub was not installed successfuly in PReP')


    def setup(self, layout):
        """
        Configures the partitioning layout desired and execute it.

        Args:
            layout (dict): dict with the full partitioning layout
        """
        self.__logger.info('Partitioner setup')
        self.__partitions = layout['partitions']
        self.__vgs = layout['vgs']
        self.__lvs = layout['lvs']

        # log the basic parititon layout
        self.__logger.info('\tPartitions:')
        self.__logger.info(pprint.pformat(self.__partitions,
                                          indent=2,
                                          width=40))

        # log the vgs to be created
        self.__logger.info('\tVGs:')
        self.__logger.info(pprint.pformat(self.__vgs,
                                          indent=2,
                                          width=40))

        # log the lvs to be create
        self.__logger.info('\tLVs:')
        self.__logger.info(pprint.pformat(self.__lvs,
                                          indent=2,
                                          width=40))

        # do partitioning
        partitions = self.setupPartitions()

        # do LVM
        self.setupLVM(partitions)


    def getPartitionLayout(self):
        """
        Get the layout current applied in BLivet but not yet commited
            in disk.

        Returns:
            str: partition layout
        """
        return str(self.__storage)


    def commit(self, teardown=True):
        """
        Commits the Blivet queue in disk.

        Args:
            teardown (bool): True to teardown devices
        """
        try:
            self.__storage.commit(teardown)

        except:
            raise


    def setupPartitions(self):
        """
        Creates and returns all partitions designed by the client. The
            strucutre returned is a dictionary with a string name that
            user defines. This will be used to find the partition
            easily in case it becomes a physical volume or a RAID
            array.

        Returns:
            dict: dict of Blivet:PartitionDevice by name
        """
        partitions = {}

        # cache for the disks to be processed
        disks = {}

        for partition in self.__partitions:

            # disk is not cached
            if partition['disk'].name not in disks:

                # check if disk exists
                tmp = self.__storage.getDisk(partition['disk'].name)
                if tmp is None:
                    raise PKVMError("072")

                # clear current partitions in disk
                self.__storage.clearDevice(tmp)

                # set GPT label
                self.__storage.setDiskLabel(tmp)

                # cache it
                disks[partition['disk'].name] = tmp

            # get disk from cache
            disk = disks[partition['disk'].name]

            # add a new partition in the disk
            part = self.__storage.addPartition(disk,
                                               partition['name'],
                                               partition['size'],
                                               partition['fsType'],
                                               partition['boot'],
                                               partition['type'],
                                               partition['weight'])

            # add the partition in the list
            partitions[partition['name']] = part

            if partition['mount'] is not None:
                self.__mountpartitions[partition['mount']] = part

        # allocates the layout from the devices
        self.__storage.processBlivetQueuePartitions()

        return partitions


    def setupLVM(self, partitions):
        """
        Creates the LVM layout using PVs defined in partitions.

        Args:
            partitions (dict):  dict of Blivet:PartitionDevice
        """
        # process volume group layout
        vgs = {}
        for vg in self.__vgs:

            # get the partition PV device that user wants to use in VG
            pvs = []
            for pv in vg['pvs']:

                pvDevice = None

                # user wants a PV but this name was not defined when creating
                # the partition, so we will look for it. This can happen if
                # user wants to create a VG using an existing PV, then it
                # should come as devicePartitionsNumber, like sda5.
                if pv not in partitions:

                    pvDevice = self.__storage.getDisk(pv)

                    # pv desired was not found anyway
                    if pvDevice is None:
                        self.__logger.critical('PV %s was not found', pv)
                        raise PKVMError("012")

                # pv name was found as a new parititon
                else:
                    pvDevice = partitions[pv]

                pvs.append(pvDevice)

            # create the volume group
            volume = self.__storage.newVG(pvs, vg['name'])

            # add in partitioner structure
            vgs[vg['name']] = volume

        # process logical volume layout
        for lv in self.__lvs:

            # user wants a VG but the name was not defined when creating
            # the volume groups
            if lv['vg'] not in vgs:
                raise PKVMError("013")

            lvdev = self.__storage.newLV(vgs[lv['vg']],
                                         lv['name'],
                                         lv['size'],
                                         lv['fsType'],
                                         lv['grow'])

            if lv['mount'] is not None:
                self.__mountpartitions[lv['mount']] = lvdev

        self.setTargetLVMInfo(vgs.values())


    def formatPartition(self, device, pformat):
        """
        Wipes out the current format in the partition and creates
        a new one as specified.

        Args:
            device (Blivet:DiskDevice): partition to be formated
            pformat (str): format type ('ext3', 'ext4', 'lvmpv', etc)
        """
        # format the disk
        self.__storage.formatPartition(device, pformat)
        self.__storage.processBlivetQueuePartitions()


    def setTargetLVMInfo(self, targetSystemVGs):
        """
        Sets the LVM configuration of the target

        Args:
            targetSystemVGs ([blivet.devices.LVMVolumeGroup]):
                volume groups of the target system
        """
        targetSystemPVs = []
        targetSystemLVs = []

        for vg in targetSystemVGs:
            for pv in vg.pvs:
                targetSystemPVs.append(pv)

            for lv in vg.lvs:
                lvNameWithIBMPrefix = lv.name.replace(vg.name + '-', '')
                lvName = lvNameWithIBMPrefix.replace(LV_PREFIX, '')

                if lvName == PKVM2_LV_LOG:
                    lvName = LV_VAR
                if lvName == PKVM2_LV_ROOT:
                    lvName = LV_SYSTEM

                if lvName in LVS_METADATA.keys():
                    # This check is to ensure we don't include any
                    # logical volumes created by the user.
                    targetSystemLVs.append(lv)

        self.targetLVMInfo = {'pvs' : targetSystemPVs,
                              'vgs' : targetSystemVGs,
                              'lvs' : targetSystemLVs}


    def setSystemVGName(self, vgName, assertUnique):
        """
        Sets the name of the volume group where PowerKVM is/will be installed.

        Args:
            vgName (str): volume group name
            assertUnique (bool): use an unique name based on vgName
        """
        if assertUnique:
            self.systemVGName = self.getUniqueVGName(vgName)
        else:
            self.systemVGName = vgName


    def isPrep(self, partition):
        """
        Checks if the partition is a PReP partition

        Args:
            partition (blivet.device.PartitionDevice): the partition to check

        Returns:
            bool: whether partition is a PReP partition
        """

        return (partition.partedPartition.name == "ibmpkvm_prep" or
                partition.format.type == "prepboot" or
                (partition.partedPartition.number == 1 and partition.size == Size(PREP_SIZE)))

    def isBoot(self, partition):
        """
        Checks if the partition is a boot partition

        Args:
            partition (blivet.device.PartitionDevice): the partition to check

        Returns:
            bool: whether partition is a boot partition
        """

        return (partition.partedPartition.name == "ibmpkvm_boot" or
                (partition.partedPartition.number == 2 and partition.size == Size(BOOT_SIZE)))

    def getUniqueVGName(self, suggestedName):
        """
        Returns an unique volume group name based on the chosen name.

        Args:
            suggestedName (str): Volume group name to be checked.

        Returns:
            str: Unique volume group name appended with digits if necessary.

        Raises:
            PKVMError: in case we can't find an unique name
        """
        names = self.__storage.getBlivetDeviceNames() + self.__storage.getVGNames()
        names = list(set(names))
        newName = suggestedName

        self.__logger.debug("Suggested VG name: %s", suggestedName)
        for i in range(100):
            if newName in names:
                newName = "%s%02d" % (suggestedName, i)
            else:
                self.__logger.debug("Using VG name: %s", newName)
                return newName
        raise PKVMError("105")
