#======
# Copyright (c) IBM Corp. 2011, 2015.
# All rights reserved. This program is made available
# under the terms of the Eclipse Public License v1.0
# which accompanies this distribution, and is available at
# http://www.eclipse.org/legal/epl-v10.html
#=======


"""This script handles the storage of PowerKVM."""


#
# IMPORTS
#

import logging
import parted
import re
import traceback
import subprocess

import blivet
from blivet import errors
from blivet.devicelibs import mpath
from blivet.errors import LVMError
from blivet.formats import getFormat
from blivet.size import Size
from pykickstart.version import makeVersion

from controller.pkvmerror import PKVMError


#
# CONSTANTS
#
PARTITION_TYPES = {
    'primary'  : parted.PARTITION_NORMAL,
    'extended' : parted.PARTITION_EXTENDED,
    'logical'  : parted.PARTITION_LOGICAL
}


#
# CODE
#
class Storage(object):
    """
    Wraps essential blivet methods in PowerKVM.
    """

    def __init__(self):
        """
        Constructs a storage object for PowerKVM to interact with
            Blivet methods.
        """
        self.__logger = logging.getLogger(__name__)
        self.__blivet = blivet.Blivet()
        self.ignoredDisks = []

        self.scanDisks()

    def _getBlivet(self):
        """
        Returns the Blivet object for arbitrary handling.

        Returns:
            Blivet: Blivet object
        """
        return self.__blivet


    def __repr__(self):
        """
        Creates a string representation of the current queue of
            partitioning commands.

        Returns:
            str: object string representation
        """
        actions = ''
        for action in self.__blivet.devicetree.findActions():
            actions += '%s\n' % str(action)[6:-8]

        return actions


    def getSystemInformation(self):
        """
        Returns system information about devices and partitions.

        IMPORTANT: the structure might not reflect the current system
            state because changes in blivet's queue will be reflected
            - even if it's not yet commited. If you want to be sure to
            get the current system state, call 'scanDisks' just
            before this one.

        Returns:
            dict: devices and partitions data
        """
        data = {}

        data['devices'] = self.__blivet.devicetree.devices
        data['partitions'] = self.partitionsInfo()
        data['disks'] = self.diskInfo()
        data['lvm'] = self.lvmInfo()
        data['raid'] = self.raidInfo()

        return data

    def scanDisks(self):
        """
        Scans disks and loads disk structure to blivet's device tree
        """
        while True:
            try:
                if not self.__blivet.ksdata:
                    self.__blivet.ksdata = makeVersion()

                self.__blivet.reset()
            except errors.DeviceTreeError as e:
                # while scanning a disk, blivet raises an exception in
                # case the disk's partition table is corrupt (as
                # reported by parted) but it can still detect the
                # existence of a partition (via udev). we are going to
                # ignore these disks and retry.

                errorMsgRegex = ("(corrupt gpt disklabel on disk|"
                                 "failed to scan disk) (.*)")
                match = re.match(errorMsgRegex, e.message)
                if match:
                    self._ignoreDevice(match.group(2))
                    continue
                else:
                    raise PKVMError("015")
            else:
                break

    def _ignoreDevice(self, deviceName):
        """
        Hide a device from blivet's device tree. The device will be
        ignored in the next call to blivet.reset().

        Args:
           deviceName (str): the name of a blivet.StorageDevice. In
           the context of the installer, this device should be a disk
           or a multipath device.
        """
        # TODO: Volume groups of an ignored device are invisible to
        # blivet and this will cause it to raise and exception if the user gives
        # us a new VG name equal to the old one.

        device = self.__blivet.devicetree.getDeviceByName(deviceName)

        if device:
            # ignore the device's partitions.
            self.ignoredDisks.append(deviceName)

            # ignore the device itself and its parents. In the case of
            # a multipath device, parents are the disks of that
            # path. A disk will have an empty list as parents.
            parentNames = [disk.name for disk in device.parents]
            self.ignoredDisks.extend(parentNames)

            self.__blivet.ksdata.ignoredisk.ignoredisk = self.ignoredDisks

    def setTargetDevices(self, devices):
        """
        Mark devices to be used. All other devices will be
        ignored. This is roughly equivalent of calling _ignoreDevice()
        for all devices not in 'devices'.

        Args:
           devices ([str]): device names. Devices could be disks,
               multipath devices, RAID arrays.
        """
        self.__blivet.ksdata.ignoredisk.onlyuse = devices
        # need to re-scan the disks for these changes to take effect.
        self.scanDisks()

    def getDisks(self):
        """
        Returns all disks found in the system.

        Returns:
            [Blivet.DiskDevice]: list of disks
        """
        disks = []
        for disk in self.__blivet.disks:

            # ignore mpath slaves
            if mpath.is_multipath_member(disk.path):
                continue
            disks.append(disk)

        return disks


    def getDisk(self, diskId):
        """
        Returns a Blivet Disk object based on a device
            identifier. This method always returns the master device
            in case of multipath. Then, if the id or name links to a
            mpath slave this will be returned the master.

        Args:
            diskId (str): disk identifier. This idenfier could be any
                of:
                    disk name  (sda);
                    disk id    (wwn-0x5000c5002f301765)
                    disk path  (pci-0005:00:01.0-scsi-0:255:0:0)
                    disk label (POWERKVM_DVD)

        Returns:
            Blivet.DiskDevice: Blivet disk object, None if disk not
                found
        """
        # resolve the disk name to use with blivet
        resolvedDiskId = blivet.udev.resolve_devspec(diskId)

        # if the name cannot be resolved, try to resolve by looking for
        # any device finishing by the id passed
        if resolvedDiskId is None:
            resolvedDiskId = blivet.udev.resolve_glob('*%s' % diskId)

            # name cannot be resolved to a single device or cannot be
            # resolve at all, we cannot them use this id. The client
            # must provide a unique way to find the device
            if len(resolvedDiskId) < 1:
                self.__logger.debug('Device %s was not found', str(diskId))
                return None

            resolvedDiskId = resolvedDiskId[0]

        # get the blivet device based on what we found
        disk = self.__blivet.devicetree.getDeviceByName(resolvedDiskId)

        # if device is a mpath slave, return the master
        try:
            if disk.format.type == 'multipath_member':
                disk = self.__blivet.devicetree.getChildren(disk)[0]
        except AttributeError:
            return None

        return disk


    def setDiskLabel(self, disk, disklabel='gpt'):
        """
        Sets the disk partition label. Is no label is passed or the
            label passed is not supported by the platform, Blivet will
            pick the best option possible.

        Args:
            disk (Blivet.DiskDevice): Blivet disk object
            disklabel (str): partitioning layout (msdos, gpt, etc)
        """
        self.__logger.info('Set %s to label %s', disk.path, disklabel)

        platform = blivet.platform.platform

        # disklabel is not available in the current platform, use
        # the best option
        if disklabel not in platform.diskLabelTypes:
            self.__logger.warn('label %s is not valid', disklabel)

            disklabel = platform.bestDiskLabelType(disk)
            self.__logger.warn('storage will set the label %s for it',
                               disklabel)

        # create a new disklabel on the disk
        newLabel = blivet.getFormat("disklabel", device=disk.path,
                                    labelType=disklabel)
        labelAction = blivet.ActionCreateFormat(disk, fmt=newLabel)

        try:
            self.__blivet.devicetree.registerAction(labelAction)

        # catch errors when registering the action
        except errors.DeviceTreeError, e:
            raise PKVMError("047")
    # setDiskLabel()

    def getPartitions(self, disk):
        """
        Returns all partitions found in a given disk.

        Args:
            disk (Blivet.DiskDevice): Blivet disk object

        Returns:
            [Blivet.PartitionDevice]: list of partitions
        """
        partitions = []
        devices = self.__blivet.deviceDeps(disk)

        for device in devices:

            # add only leaves
            if device.type == 'partition':
                partitions.append(device)

        return partitions


    def addPartition(self, disk, name, size, pformat, bootable,
                     ptype=None, number=0):
        """
        Creates a new partition in a given disk.

            disk (Blivet:DiskDevice): Blivet disk object
            name (str): partition name
            size ( str): partition size, i.e: 10GiB, 800MB,
                ALL_FREESPACE
            pformat (str): format (ext4, prepboot, lvmpv, ...)
            bootable (bool): True for a bootable partition
            ptype (str): partition type (primary, logical, extended)
                for non GPT.
            number (int): partition number desired (0 will be sorted)

        Returns:
            Blivet:PartitionDevice: partition to be later used (LVM,
                RAID)
        """
        self.__logger.info('Create partition')

        # only consider valid partition types
        partType = None
        if ptype in PARTITION_TYPES:
            partType = PARTITION_TYPES[ptype]

        part = None

        # when requesting a partition to use all the free space
        # available, we set it growable without passing the size
        grow = False
        if size == 'ALL_FREESPACE':
            grow = True
            size = None

        else:
            size = Size(size)

        try:
            # create a new partition
            part = self.__blivet.newPartition(parents=[disk],
                                              name=name,
                                              size=size,
                                              fmt=getFormat(pformat),
                                              bootable=bootable,
                                              partType=partType,
                                              weight=number,
                                              grow=grow)

        except errors.DeviceError, e:
            raise PKVMError("100")

        # schedule creation of the partition
        self.__blivet.createDevice(part)

        self.__logger.debug(
            'Added Disk: %s, Size: %s, Partition: %s, Format: %s',
            disk.name,
            str(part.size),
            part.name,
            pformat)

        return part


    def clearDevice(self, device):
        """
        Clears a given device recursively. It means that if the device
            has any partition inside all of them will be removed as
            well as any LVM or RAID that relies on this device's
            state.

        Args:
            device (Blivet:DiskDevice): Blivet disk object
        """
        try:
            self.__blivet.recursiveRemove(device)

        # catch device inconsistences
        except ValueError, e:
            raise PKVMError("095")

        # catch device error
        except errors.DeviceTreeError, e:
            raise PKVMError("009")

        except Exception, e:
            raise PKVMError("099")


    def diskInfo(self):
        """
        Returns a data structure containing information about all
            disks found in the running system.

        IMPORTANT: the partition list returned could not reflect the
            actual disks states WHEN there are actions in
            Blivet queue waiting to be processed.

        Returns:
            [Blivet.DiskDevice]: list of partitions found
        """
        return self.getDisks()


    def partitionsInfo(self):
        """
        Returns all partitions in all disks found in the system.

        Returns:
            dict: partitions per disk name
        """
        partitions = {}
        for disk in self.getDisks():
            partitions[disk.name] = self.getPartitions(disk)

        return partitions


    def newVG(self, pvs, name):
        """
        Creates a new Volume Group using a list of Physical Volumes.

        Args:
            pvs ([str]): list of PV partitions
            name (str): VG name

        Returns:
            Blivet.LVMVolumeGroupDevice: volume group created
        """
        self.__logger.info('Create Volume Group')

        vg = None

        # pvs must be a list
        if not isinstance(pvs, list):
            raise PKVMError("044")

        try:
            vg = self.__blivet.newVG(parents=pvs, name=name)
            self.__blivet.createDevice(vg)

        # catch vg inconsistences
        except ValueError, e:
            raise PKVMError("011")

        # catch device error
        except errors.DeviceTreeError, e:
            raise PKVMError("101")

        self.__logger.info('Volume Group %s create', name)

        return vg


    def newLV(self, vg, name, size, pformat, grow=True):
        """
        Creates a new Logical Volume for a given list of Volume Groups.

        Args:
            vg (Blivet.LVMVolumeGroupDevice): volume groupe
            name (str): LV name
            size (int): LV size required (1GiB, 500MB, ALL_FREESPACE)
            pformat (str): filesystem format (ext4, ext3, etc)
            grow (bool): True if LV can grow.

        Returns:
            Blivet.LVMLogicalVolumeDevice: logical Volume
        """
        self.__logger.info('Create Logical Volume')

        # when requesting a partition to use all the free space
        # available, we set it growable without passing the size
        if size == 'ALL_FREESPACE':
            grow = False
            size = vg.freeSpace

        else:
            size = Size(size)

        try:
            lv = self.__blivet.newLV(fmt_type=pformat,
                                     grow=grow,
                                     parents=[vg],
                                     name=name,
                                     size=size)

            self.__blivet.createDevice(lv)

        # catch vg inconsistences
        except ValueError, e:
            raise PKVMError("045")

        # catch device error
        except errors.DeviceTreeError, e:
            raise PKVMError("102")

        self.__logger.info('Logical Volume %s created', name)

        return lv



    def lvmInfo(self):
        """
        Returns a data structure filled with LVM information found in
            the running system.

        Returns:
            dict: dict with information on physical volumes, volume
                groups and logical volumes.
        """
        return {'pvs' : self.__blivet.pvs,
                'vgs' : self.__blivet.vgs,
                'lvs' : self.__blivet.lvs
               }


    def newMDRaid(self, name, parents, level, numSpares, pformat):
        """
        Creates a MDRAID using a list of RAID partitions as array.

        Args:
            name (str): name of the Raid (/dev/md/<name>)
            parents ([Disk]): parentes of this device
            level (str): RAID level (raid0, raid1, raid5, ...)
            totalDevices (int): number of spares devices
            pformat (blivet.formats): filesystem format (ext4, ext3, etc)
        """
        self.__logger.info('Create RAID array')

        try:
            totalDevices = len(parents)
            membersTotal = totalDevices - numSpares

            mdRaid = self.__blivet.newMDArray(name=name,
                                              parents=parents,
                                              level=level,
                                              memberDevices=membersTotal,
                                              totalDevices=totalDevices,
                                              fmt=getFormat(pformat))

            self.__blivet.createDevice(mdRaid)

        # catch device inconsistences
        except ValueError, e:
            raise PKVMError("043")

        # catch device error
        except errors.DeviceTreeError, e:
            raise PKVMError("014")

        except Exception, e:
            raise PKVMError("103")


    def raidInfo(self):
        """
        Returns a data structure filled with RAID information found in
        the running system.

        Returns:
            dict: dict in the following format: { RAID_NAME: Size(),
                parents:[],
                parentInfo: {
                    <device>: {size:Size(<size>),
                    path:<path>,
                    parents:<parents>
                   }
                }
         }
        """

        raidInfo = {}

        for member in self.__blivet.mdarrays:

            parentList = []
            parentListMember = []
            parentInfo = {}

            for parent in member.parents:
                parentList.append(parent.name)

                for parentMember in parent.parents:
                    parentListMember.append(parentMember.name)

                parentInfo[parent.name] = {'size': parent.size,
                                           'path': parent.path,
                                           'parents': parentListMember,
                                          }

            raidInfo[member.name] = {'size': member.size,
                                     'parents': parentList,
                                     'parentInfo': parentInfo
                                    }


        return raidInfo


    def formatPartition(self, device, pformat):
        """
        Wipes out the current format in the partition and creates
        a new one as specified.

        Args:
            device (Blivet.DiskDevice): partition to be formated
            pformat (str): format type ('ext3', 'ext4', 'lvmpv', etc)
        """
        self.__logger.info('Formating %s to %s', device.path, pformat)

        # format the disk
        self.__blivet.formatDevice(device, getFormat(pformat))


    def processBlivetQueuePartitions(self):
        """
        Allocates and grows partitions. After this method call, the
        storage queue is ready to be commit.
        """
        self.__logger.info('Processing the queue')

        try:
            blivet.partitioning.doPartitioning(self.__blivet)

        except errors.PartitioningError, e:
            raise PKVMError("010")

        # catch everthing else
        except Exception, e:
            raise PKVMError("104")


    def commit(self, teardown=True):
        """
        Commits all changes requested.

        Args:
            teardown (bool): True to teardown devices
        """
        self.__logger.info('Executing the following actions - %s',
                           str(self))

        # process the Blivet queue
        try:
            if teardown:
                self.__logger.debug('Tearing down all devices')
                self.__blivet.devicetree.teardownAll()

            self.__blivet.devicetree.processActions()

        # LVM issues: just raise it up
        except LVMError:
            raise PKVMError("097")

        # catch device tree issues
        except errors.DeviceTreeError, e:
            raise PKVMError("070")

        # catch everthing else
        except Exception, e:
            raise PKVMError("098")

    def getVGNames(self):
        """
        Return the names of all existing volume groups as reported by pvs

        Returns:
            [str]: volume group names
        """
        cmd = "pvs --unquoted --noheadings -ovg_name"

        try:
            self.__logger.debug("Running command: %s", cmd)
            sp = subprocess.Popen(cmd, shell=True, stdout=subprocess.PIPE,
                                  stderr=subprocess.PIPE, close_fds=True)
            (out, err) = sp.communicate()
            rc = sp.returncode

            if err:
                self.__logger.warn(err)
            if rc != 0:
                self.__logger.warn("Could not get volume group names. Return code: %s", rc)
                return []
        except OSError as e:
            self.__logger.warn("Could not get volume group names: %s", str(e))
            return []
        else:
            return out.strip().split()

    def getBlivetDeviceNames(self):
        """
        Return the device names known by Blivet

        Returns:
            [str]: blivet device names
        """
        return self.__blivet.names
