#======
# Copyright (c) IBM Corp. 2011, 2015.
# All rights reserved. This program is made available
# under the terms of the Eclipse Public License v1.0
# which accompanies this distribution, and is available at
# http://www.eclipse.org/legal/epl-v10.html
#=======


#
# IMPORTS
#
import socket
import time

from snack import *

from .utils import addFieldToGrid, addJoinedFieldsToGrid
from .utils import genDateStr, genTimeStr
from modules.i18n.i18n import getLangConf
from viewer.__data__ import BACK
from viewer.__data__ import DATE_DAY_LABEL
from viewer.__data__ import DATE_LABEL
from viewer.__data__ import DATE_MONTH_LABEL
from viewer.__data__ import DATE_TIME_SETUP
from viewer.__data__ import DATE_YEAR_LABEL
from viewer.__data__ import ERROR_DATETIME_FORMAT
from viewer.__data__ import ERROR_DATETIME_FORMAT_MSG
from viewer.__data__ import OK
from viewer.__data__ import TIME_HOUR_LABEL
from viewer.__data__ import TIME_LABEL
from viewer.__data__ import TIME_MINUTE_LABEL
from viewer.__data__ import TIME_SECOND_LABEL

#
# CONSTANTS
#


#
# CODE
#
class DateTimeSetup:
    """
    Represents the Datetime configuration screen
    """

    def __init__(self, screen, params):
        """
        Constructor

        Args:
            screen (SnackScreen): SnackScreen instance
            params (dict): screen input parameters
        """
        self.__screen = screen

        langConf = getLangConf()
        self.__contentGrid = Grid(1, 2)

        self.__listYear   = Listbox(1, scroll=0, width=4, returnExit=1)
        self.__listMonth  = Listbox(1, scroll=0, width=2, returnExit=1)
        self.__listDay    = Listbox(1, scroll=0, width=2, returnExit=1)
        self.__listHour   = Listbox(1, scroll=0, width=2, returnExit=1)
        self.__listMinute = Listbox(1, scroll=0, width=2, returnExit=1)
        self.__listSecond = Listbox(1, scroll=0, width=2, returnExit=1)

        dateLabel = Label(DATE_LABEL.localize() + ' (' + \
                          genDateStr(langConf['datetime'],
                                     DATE_YEAR_LABEL.localize(),
                                     DATE_MONTH_LABEL.localize(),
                                     DATE_DAY_LABEL.localize()) + \
                          '): ')

        timeLabel = Label(TIME_LABEL.localize() + ' (' + \
                          genTimeStr(TIME_HOUR_LABEL.localize(),
                                     TIME_MINUTE_LABEL.localize(),
                                     TIME_SECOND_LABEL.localize()) + \
                          '): ')

        dateGrid = Grid(6, 1)
        timeGrid = Grid(6, 1)

        addFieldToGrid(dateGrid, dateLabel, 0, 0)
        addFieldToGrid(timeGrid, timeLabel, 0, 0)

        def dateTagToField(fieldTag):
            """
            Reads a date tag ('Y' or 'M' or 'D') and returns the
                corresponding field.

            Args:
                fieldTag (string):

            Returns:
                ListBox:
            """
            if fieldTag == 'Y':
                return self.__listYear
            elif fieldTag == 'M':
                return self.__listMonth
            else:
                return self.__listDay

        dateFormat = langConf['datetime'].split()

        addJoinedFieldsToGrid(dateGrid, '/', map(dateTagToField, dateFormat),
                              1, 0)
        addJoinedFieldsToGrid(timeGrid, ':', [self.__listHour,
                                              self.__listMinute,
                                              self.__listSecond], 1, 0)

        addFieldToGrid(self.__contentGrid, dateGrid, 0, 0)
        addFieldToGrid(self.__contentGrid, timeGrid, 0, 1)

        self.__buttonsBar = ButtonBar(self.__screen, [(OK.localize(),"ok"),
            (BACK.localize(), "back")])

        self.__grid = GridForm(self.__screen, DATE_TIME_SETUP.localize(), 1, 2)
        self.__grid.add(self.__contentGrid, 0, 0, (0,0,0,1))
        self.__grid.add(self.__buttonsBar, 0, 1)

    def run(self):
        """
        Draws the screen

        Returns:
            dict: screen output parameters. Valid keys:
                pressedButton str pressed button ID
                datetime date/time selected by user
        """

        localtime = time.localtime()

        # populate listboxes
        for i in range(2014, 2035):
            self.__listYear.append(str(i), str(i))

        if localtime.tm_year < 2014:
            self.__listYear.setCurrent("2014")
        else:
            self.__listYear.setCurrent(str(localtime.tm_year))

        for month in DateTimeSetup.__strRange(1, 13):
            self.__listMonth.append(month, month)

        self.__listMonth.setCurrent(str(localtime.tm_mon).zfill(2))

        for day in DateTimeSetup.__strRange(1, 32):
            self.__listDay.append(day, day)

        self.__listDay.setCurrent(str(localtime.tm_mday).zfill(2))

        for hour in DateTimeSetup.__strRange(0, 24):
            self.__listHour.append(hour, hour)

        self.__listHour.setCurrent(str(localtime.tm_hour).zfill(2))

        for minute in DateTimeSetup.__strRange(0, 60):
            self.__listMinute.append(minute, minute)

        self.__listMinute.setCurrent(str(localtime.tm_min).zfill(2))

        for second in DateTimeSetup.__strRange(0, 60):
            self.__listSecond.append(second, second)

        self.__listSecond.setCurrent(str(localtime.tm_sec).zfill(2))

        result = self.__grid.run()
        self.__screen.popWindow()

        _datetime = {
                "year": self.__listYear.current(),
                "month": self.__listMonth.current(),
                "day": self.__listDay.current(),
                "hour": self.__listHour.current(),
                "minute": self.__listMinute.current(),
                "second": self.__listSecond.current(),
                }

        return {"pressedButton": self.__buttonsBar.buttonPressed(result),
                "datetime": _datetime}

    def showErrorWrongFormat(self, str_date):
        """
        Displays an error about the datetime format
        """
        ButtonChoiceWindow(self.__screen, ERROR_DATETIME_FORMAT.localize(),
                           ERROR_DATETIME_FORMAT_MSG.localize() % str_date,
                           buttons=[(OK.localize(), 'ok'), (BACK.localize(),
                               'back')],
                           width = 50)

    @staticmethod
    def __strRange(start, end, fill = 2):
        """
        Similar to python's range, but returns a list of strings with
        leading zeroes. The 'fill' parameter represents the number of
        characters the number string must at least have. (1 would turn
        into '01').  Uses 1 as step.

        Args:
            start (int): start of range
            end (int): end of range (not closed interval)
            fill (int): mininum number of characters a number string
                should have.


        Returns:
            [string]: something like ['09', '10']

        """
        return [str(num).zfill(fill) for num in range(start, end, 1)]
